/*
//  File  : pooling.cpp
//  (c)HOOPOE ProGroup 2011
*/

/*
// C= Programming Sample
*/

/// Copyright string
char* copyright = "(c)HOOPOE ProGroup 2011";

#include <iostream>
#include <vector>

#include <C=.h>

using namespace std;

class CPool
{
    vector<string> input;
    vector<string> store;

    vector<string>* output;

    /// thread ID map
    void* id[3];

    /// thread count
    volatile int thread_no;
    /// destruction flag
    volatile int destroy;

public:

    CPool() : thread_no(0), destroy(0)
    {
        /// ensure there're no surprizes when accessing the class from multiple threads
        CPool* shared obj = this;

        parallel(3)
        {
            /// create 3 threads and save their global IDs
            obj->id[pix()] = pid();

            serial()
            {
                obj->thread_no++;
                break();
            }

            switch(pix())
            {
            case 0:

                /// wait for other threads to initialize
                while(obj->thread_no != pno()) wait(0);

                break;

            case 1:

                /// process requests
                for(;;)
                {
                    size_t reqno;

                    wait();

                    if(obj->destroy)
                    {
                        break;
                    }
                    while(reqno = obj->input.size())
                    {
                        /// create a private pool to process outstanding requests in parallel
                        parallel(reqno)
                        {
                            string s;

                            /// sample a request
                            serial(&obj->input)
                            {
                                s = obj->input[pix()];
                                break();
                            }
                            /// process data
                            for(string::iterator i = s.begin(); i < s.end(); i++)
                            {
                                if(*i >= 'A' && *i <= 'Z')
                                {
                                    *i |= 0x20;
                                }
                            }
                            /// store the results
                            serial(&obj->store)
                            {
                                obj->store.push_back(s);
                                break();
                            }
                            barrier();
                        }
                        /// remove processed requests
                        serial(&obj->input)
                        {
                            obj->input.erase(obj->input.begin(), obj->input.begin() + reqno);
                            break();
                        }
                    }
                }
                /// indicate the destruction has completed
                serial()
                {
                    obj->destroy++;
                    break();
                }
                break;

            case 2:

                /// form replies
                for(;;)
                {
                    size_t reqno;

                    wait();

                    if(obj->destroy)
                    {
                        break;
                    }
                    /// fetch ready data (if any)
                    while(reqno = obj->store.size())
                    {
                        /// create a private pool to form the reply in parallel
                        parallel(reqno)
                        {
                            string s;

                            /// sample a data element
                            serial(&obj->store)
                            {
                                s = obj->store[pix()];
                                break();
                            }
                            /// form the reply
                            serial(obj->output)
                            {
                                obj->output->push_back(s);
                                break();
                            }
                            barrier();
                        }
                        /// remove processed elements
                        serial(&obj->store)
                        {
                            obj->store.erase(obj->store.begin(), obj->store.begin() + reqno);
                            break();
                        }
                    }
                    /// signal the reply is ready
                    wake(obj->id[0]);
                }
                /// indicate the destruction has completed
                serial()
                {
                    obj->destroy++;
                    break();
                }
                break;
            }
            break();
        }
    }

    ~CPool()
    {
        destroy = 1;

        for(int i = 1; i < thread_no; i++)
        {
            wake(id[i]);
        }
        while(destroy != thread_no) wait(0);
    }

    void placeRequest(string req)
    {
        serial(&input)
        {
            input.push_back(req);
            break();
        }
        wake(id[1]);
    }

    void fetchReply(vector<string>* rep)
    {
        output = rep;
        wait(0);
        wake(id[2]);
        wait();
    }
};

void main()
{
    /// TODO: Illustrate "private thread pools":
    ///
    ///     * on-demand high-priority processing
    ///     * no CPU oversubscription
    ///     * multiple pools can operate in parallel
    ///     * tailored to the logic of your particular class
    ///

    CPool pool;

    vector<string> inlines;

    inlines.push_back("HeLLo, wOrLd!");
    inlines.push_back("MessAgE CASE ConveRsion");
    inlines.push_back("viA PARAllEl PoOLs");

    for(vector<string>::iterator i = inlines.begin(); i < inlines.end(); i++)
    {
        pool.placeRequest(*i);
    }
    for(vector<string>::iterator i = inlines.begin(); i < inlines.end(); i++)
    {
        cout << i->c_str() << endl;
    }
    cout << endl;

    vector<string> outlines;

    pool.fetchReply(&outlines);

    for(vector<string>::iterator i = outlines.begin(); i < outlines.end(); i++)
    {
        cout << i->c_str() << endl;
    }
    cout << endl;
}
