/*
//  File  : shaker.c
//  (c)HOOPOE ProGroup 2011
*/

/*
// C= Programming Sample
*/

/// Copyright string
char* copyright = "(c)HOOPOE ProGroup 2011";

#include "internal.h"
#include "C=.h"

/// 2D vector type
typedef struct
{
    int x;
    int y;

} vec2d_t;

/// window movement vector
vec2d_t vw = {0, 0};

/// particles
#define MAXPART 100

/// center coordinates
vec2d_t* pa;
vec2d_t* pb;

/// velocity vectors
vec2d_t* va;
vec2d_t* vb;

/// main window handle
HWND hwndMain;

/// previous window coordinates
RECT wprev;

/// window messaging callback
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    RECT wcurr;

    switch (uMsg) 
    {
    case WM_DESTROY:
        /// unregister all windows
        PostQuitMessage(0);
        return 0;

    case WM_MOVE:

        GetWindowRect(hwnd, &wcurr);

        /// synchronize access to the window movement vector
        serial(&vw)
        {
            vw.x = wprev.left - wcurr.left;
            vw.y = wprev.top - wcurr.top;

            wprev = wcurr;

            break();
        }

    default:
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    } 
    return 0; 
}

/// main window creation
int CreateMainWindow(HINSTANCE hinstance)
{
    WNDCLASS wc;

    /// register the main window class
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC) MainWndProc; 
    wc.cbClsExtra = 0; 
    wc.cbWndExtra = 0; 
    wc.hInstance = hinstance; 
    wc.hIcon = LoadIcon(hinstance, (LPCSTR)IDI_ICON1); 
    wc.hCursor = LoadCursor(NULL, IDC_ARROW); 
    wc.hbrBackground = GetStockObject(DKGRAY_BRUSH); 
    wc.lpszMenuName =  "MainMenu"; 
    wc.lpszClassName = "MainWindowClass"; 
 
    if(!RegisterClass(&wc))
    {
        return -1; 
    }
    /// create the main window
    hwndMain = CreateWindowEx( 
        0,                                  // no extended styles           
        "MainWindowClass",                  // class name                   
        "Shaker /// C= Programming Sample",    // window name                  
        WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX,
        CW_USEDEFAULT,                      // default horizontal position  
        CW_USEDEFAULT,                      // default vertical position    
        CW_USEDEFAULT,                      // default width                
        CW_USEDEFAULT,                      // default height               
        (HWND)NULL,                         // no parent or owner window    
        (HMENU)NULL,                        // class menu used              
        hinstance,                          // instance handle              
        NULL);                              // no window creation data      

    if(!hwndMain)
    {
        return -1;
    }

    GetWindowRect(hwndMain, &wprev);
    MoveWindow(hwndMain, wprev.left, wprev.top, 500, 500, 1);

    ShowWindow(hwndMain, SW_SHOWNORMAL);
    UpdateWindow(hwndMain); 

    return 0;
}

/// entry point
int APIENTRY WinMain(HINSTANCE hinstance, HINSTANCE hinstPrev, LPSTR lpCmdLine, int nCmdShow)
{
    /// TODO: Write a whole GUI application in C=
    ///       with message processing, timers, and data-parallel computations
    /// create a window
    /// handle messages
    /// compute a window movement vector
    /// shake particles inside the window
    /// draw the window 25 times a second

    void* id_store[3];              /// parallel IDs for thread interaction
    void** shared id = id_store;    /// access pointer

    /// create the main window
    if(CreateMainWindow(hinstance))
    {
        return -1;
    }

    /// do message processing, painting, and scene computation in parallel
    parallel(3)
    {
        /// preserve global thread IDs
        id[pix()] = pid();

        switch(pix())
        {
        case 0:

            /// start the message loop in thread 0
            for(;;)
            {
                MSG msg;
                BOOL bRet;

                if(!(bRet = GetMessageA(&msg, NULL, 0, 0)))
                {
                    break;
                }
                if (bRet == -1)
                {
                    break;
                }
                else
                {
                    TranslateMessage(&msg); 
                    DispatchMessage(&msg);
                }
            }
            ExitProcess(0);

        case 1:

            /// implement a timer to signal to thread 2 each 1/25th of a second
            for(;;)
            {
                /// wait for 40 ms
                wait(40);

                /// signal to thread 1
                wake(id[2]);
            }

        case 2:

            /// create particles and initialize coordinates and velocities
            {
                int i;
                RECT client;

                /// allocate storage for particles and velocities
                pa = malloc(sizeof(vec2d_t) * MAXPART);
                pb = malloc(sizeof(vec2d_t) * MAXPART);
                va = malloc(sizeof(vec2d_t) * MAXPART);
                vb = malloc(sizeof(vec2d_t) * MAXPART);

                GetClientRect(hwndMain, &client);

                /// initialize the positions and velocities
                for(i = 0; i < MAXPART; i++)
                {
                    pa[i].x = rand() % client.right;
                    pa[i].y = rand() % client.bottom;

                    va[i].x = 8 - (rand() % 16);
                    va[i].y = 8 - (rand() % 16);
                }
            }
            /// start the computations in thread 2 (if signaled from thread 1)
            for(;;)
            {
                int i;
                HDC hdc;
                RECT client_rect;
                RECT* shared client = &client_rect;
                int vecx;
                int vecy;
                vec2d_t* tmp;

                /// clear oustanding signals (if any)
                wait(0);
                /// wait infinitely for a new signal
                wait();

                GetClientRect(hwndMain, client);

                /// sample the current window vector (in a thread-safe manner)
                serial(&vw)
                {
                    vecx = vw.x;
                    vecy = vw.y;

                    vw.x = vw.y = 0;

                    break();
                }

                /// compute new positions in parallel (on all available processors)
                parallel()
                {
                    int i = CPX_RANGE_START(MAXPART);
                    int e = CPX_RANGE_END(MAXPART);
                    int j, d1, d2;
                    vec2d_t a, b;

                    /// process the array of particles (decomposed statically)
                    for(; i < e; i++)
                    {
                        /// detect collisions and copy new positions and velocities to pb[] and vb[]
                        for(j = 0; j < MAXPART; j++)
                        {
                            /// copy previous velocities by default
                            vb[i] = va[i];

                            /// compute the distance between two balls in the middle of their trajectory
                            a.x = pa[i].x + va[i].x / 2;
                            a.y = pa[i].y + va[i].y / 2;
                            b.x = pa[j].x + va[j].x / 2;
                            b.y = pa[j].y + va[j].y / 2;

                            a.x -= b.x;
                            a.y -= b.y;

                            d1 = a.x * a.x + a.y * a.y;

                            /// compute the distance between the two balls at the beginning of their trajectory
                            a.x = pa[i].x;
                            a.y = pa[i].y;
                            b.x = pa[j].x;
                            b.y = pa[j].y;

                            a.x -= b.x;
                            a.y -= b.y;

                            d2 = a.x * a.x + a.y * a.y;

                            /// the balls approach each other...
                            if(d1 < d2)
                            {
                                /// ...and the distance between them is less than 10 points
                                if(d1 < 100)
                                {
                                    /// synchronize with other threads
                                    serial()
                                    {
                                        /// swap impulses
                                        vec2d_t tmp;
                                        tmp = va[i]; va[i] = va[j]; va[j] = tmp;
                                        vb[i] = va[i];
                                        break();
                                    }
                                }
                            }
                        }
                        /// make a move
                        pb[i].x = pa[i].x + vb[i].x + vecx;
                        pb[i].y = pa[i].y + vb[i].y + vecy;

                        /// compute border reflections
                        if(pb[i].x < 0)
                        {
                            pb[i].x = 0;
                            vb[i].x = -vb[i].x;
                        }
                        if(pb[i].x > client->right)
                        {
                            pb[i].x = client->right;
                            vb[i].x = -vb[i].x;
                        }
                        if(pb[i].y < 0)
                        {
                            pb[i].y = 0;
                            vb[i].y = -vb[i].y;
                        }
                        if(pb[i].y > client->bottom)
                        {
                            pb[i].y = client->bottom;
                            vb[i].y = -vb[i].y;
                        }
                    }
                    barrier();
                }
                /// swap the arrays of particles and velocities
                tmp = pb; pb = pa; pa = tmp;
                tmp = vb; vb = va; va = tmp;

                /// prepare for painting
                hdc = GetDC(hwndMain);

                /// clear the window
                PaintDesktop(hdc);

                /// draw the particles
                for(i = 0; i < MAXPART; i++)
                {
                    SelectObject(hdc, GetStockObject(DC_PEN));
                    SetDCPenColor(hdc, RGB(255, 150 + i, 55));

                    SelectObject(hdc, GetStockObject(DC_BRUSH));
                    SetDCBrushColor(hdc, RGB(255, 150 + i, 55));

                    Ellipse(hdc, pa[i].x - 5, pa[i].y - 5, pa[i].x + 5, pa[i].y + 5);
                }
                ReleaseDC(hwndMain, hdc);
            }
        }
        break();
    }
    return 0; 
}
